#ifndef CONSOLHWF2MAT_H
#define CONSOLHWF2MAT_H

#include "aliast.h"
#include "genfile.h"
#include "lewis_schrefler.h"
#include "van_genuchten.h"

class con_hwf2mat
{
 public:
  con_hwf2mat();    //constructor
  ~con_hwf2mat();   //destructor

  void read(XFILE *in);
  void print(FILE *out);
  double get_betas(long ipp);
  double get_cps(double t);
  double get_rhos(double t);
  double get_sw(double pw, double t, long ipp);
  double get_dsw_dpw(double pw, double t, long ipp);
  double get_dsw_dt(double pw, double t, long ipp);
  double get_kintr(double pw, double t, long ipp);
  double get_krw(double pw, double t, long ipp);

  double get_alpha(double pw, double t,long ipp);
  double get_porosity(double pw,double t,long ipp);
  double get_ks(double pw,double t,long ipp);
  double get_betasw(double pw,double t,long ipp);
  double get_rhow(double t);
  double get_kw(double pw,double t,long ipp);

  double get_muw(double t);
  double get_cpw();
  double get_betaw(double t);
  double get_rhocp(double pw,double t,long ipp);
  double get_lambdaeff(double pw,double t,long ipp);

  void matcond (matrix &d,long ri,long ci,long ipp);
  void matcap (double &c,long ri,long ci,long ipp);
  
  void matcond1d (matrix &d,long ri,long ci,long ipp);
  void matcond2d (matrix &d,long ri,long ci,long ipp);
  void matcond3d (matrix &d,long ri,long ci,long ipp);

  void matcond2 (matrix &d,long ri,long ci,long ipp);
  
  void matcond1d_2 (matrix &d,long ri,long ci,long ipp);
  void matcond2d_2 (matrix &d,long ri,long ci,long ipp);
  void matcond3d_2 (matrix &d,long ri,long ci,long ipp);

  void rhs_volume (matrix &d,long ri,long ci,long ipp);
  void rhs_volume2 (double &cc,long ri,long ci,long ipp);
  void rhs1d1 (matrix &d,long ri,long ci,long ipp);
  void rhs2d1 (matrix &d,long ri,long ci,long ipp);
  void rhs3d1 (matrix &d,long ri,long ci,long ipp);

  double transmission_transcoeff(double trc,long ri,long ci,long nn,long bc,long ipp);
  double transmission_transcoeff(double trc,long ri,long ci,long nn,long bc,long ipp,int flag);
  double transmission_nodval(double nodval,double trc2,long ri,long ci,long nn,long bc,long ipp);
  double transmission_flux(double nodval,double trc2,long ri,long ci,long nn,long bc,long ipp);

  void waterpress_check(double &pw,double t,long ipp);

  void values_correction (vector &nv);

  double get_kww(double pw,double t,long ipp);
  double get_capww(double pw,double t,long ipp);
  double get_kwt(double pw,double t,long ipp);
  double get_capwt(double pw,double t,long ipp);

  double get_ktt1(double pw,double t,long ipp);
  double get_ktt2(double pw,double t,long ipp);
  double get_captt(double pw,double t,long ipp);
  double get_ktw(double pw,double t,long ipp);
  double get_captw(double pw,double t,long ipp);

  double get_ktt2a(double pw,double t,long ipp);
  double get_ktt2c(double pw,double t,long ipp);

  double get_fw1(double pw,double t,long ipp);
  double get_ft1(double pw,double t,long ipp);
  double get_fwu(double pw, double t, long ipp);

  double get_transmission_transcoeff_ww(double pw,double t,long bc,long ipp);
  double get_transmission_transcoeff_ww(double pw,double t,long bc,long ipp,int flag);
  double get_transmission_nodval_ww(double bv,double pw,double t,long bc,long ipp);
  double get_transmission_flux_ww(double bv,double pw,double t,long bc,long ipp);

  double get_transmission_transcoeff_tt(double pw,double t,long bc,long ipp);
  double get_transmission_nodval_tt(double bv,double trr,double pw,double t,long bc,long ipp);
  double get_transmission_flux_tt(double bv,double trr,double pw,double t,long bc,long ipp);

  double get_othervalue(long compother,long ipp,double *r);
  void print_othervalue_name(FILE *out,long compother);
  void updateval (long ipp);
  void initval(long ipp);
  
  /// returns ordered dof names
  void give_dof_names(namevart *dofname, long ntm);

  double give_temperature(long ipp);
  double give_water_pressure(long ipp);
  double give_pore_pressure(long ipp);
  double give_suction(long ipp);
  double give_saturation_degree(long ipp);
  double get_w(double pw,double t,long ipp);

  /// marks required non-transport quantities
  void give_reqntq(long *antq);

  /// Lewis and Schrefler's retention curve:
  lewis_reten lewis_ret;
  /// van Genuchten's retention curve:
  van_genuchten_reten van_genuchten_ret;

 private: 
  
  heatwaterflowtype model_type;
  int compress;
  double alpha0,ks0,phi0,kintr0,betas0,rhos0,cps0,lambda0;
  double lambda1,lambda2,sr1,sr2;
  double kw0,rhow0,muw0,cpw;

  double mefel_units; //basic units for pressures = Pa (Pascals)
};  

#endif
